from pathlib import Path

from ghidrecomp import decompile, get_parser
from ghidrecomp.decompile import get_bin_output_path, gen_proj_bin_name_from_path

def test_decomplie_afd_callgraphs(shared_datadir: Path):

    parser = get_parser()

    bin_path = shared_datadir / 'afd.sys.10.0.22621.1415'

    args = parser.parse_args([f"{bin_path.absolute()}", "--callgraph-filter", "AfdRe",
                             "--filter", "AfdRe", "--callgraphs", "--skip-cache"])

    bin_proj_name = gen_proj_bin_name_from_path(bin_path)
    expected_output_path = get_bin_output_path(args.output_path, bin_proj_name)

    all_funcs, decompilations, output_path, compiler, lang_id, callgraphs = decompile(args)

    assert len(all_funcs) == 73
    assert len(decompilations) == 73
    assert output_path == expected_output_path
    assert compiler == 'visualstudio:unknown'
    assert lang_id == 'x86:LE:64:default'
    assert len(callgraphs) == 73


def test_decomplie_afd_callgraphs_cached(shared_datadir: Path):

    parser = get_parser()

    bin_path = shared_datadir / 'afd.sys.10.0.22621.1415'

    args = parser.parse_args([f"{bin_path.absolute()}", "--callgraph-filter", "AfdRe",
                             "--filter", "AfdRe", "--callgraphs"])

    bin_proj_name = gen_proj_bin_name_from_path(bin_path)
    expected_output_path = get_bin_output_path(args.output_path, bin_proj_name)

    all_funcs, decompilations, output_path, compiler, lang_id, callgraphs = decompile(args)

    assert len(all_funcs) == 73
    assert len(decompilations) == 0
    assert output_path == expected_output_path
    assert compiler == 'visualstudio:unknown'
    assert lang_id == 'x86:LE:64:default'
    assert len(callgraphs) == 0

def test_decomplie_afd_callgraphs_called_and_calling(shared_datadir: Path):

    parser = get_parser()

    bin_path = shared_datadir / 'afd.sys.10.0.22621.1415'

    args = parser.parse_args([f"{bin_path.absolute()}", "--callgraph-filter", "AfdRe",
                             "--filter", "AfdRe", "--callgraphs", "--skip-cache", "--cg-direction", "both"])

    bin_proj_name = gen_proj_bin_name_from_path(bin_path)
    expected_output_path = get_bin_output_path(args.output_path, bin_proj_name)

    all_funcs, decompilations, output_path, compiler, lang_id, callgraphs = decompile(args)

    assert len(all_funcs) == 73
    assert len(decompilations) == 73
    assert output_path == expected_output_path
    assert compiler == 'visualstudio:unknown'
    assert lang_id == 'x86:LE:64:default'
    assert len(callgraphs) == 146


import time
import base64
import zlib
import json
import sys
import re
import argparse

from typing import TYPE_CHECKING
from functools import lru_cache

# don't really limit the graph
MAX_DEPTH = sys.getrecursionlimit() - 1
MAX_PATH_LEN = 50

# needed for ghidra python vscode autocomplete
if TYPE_CHECKING:
    import ghidra
    from ghidra_builtins import *


def add_cg_args_to_parser(parser: argparse.ArgumentParser):

    group = parser.add_argument_group('Callgraph Options')
    group.add_argument('--callgraphs',  help='Generate callgraph markdown', action='store_true')
    group.add_argument('--callgraph-filter',
                       help='Only generate callgraphs for functions matching filter', default='.')
    group.add_argument('--mdd', '--max-display-depth', help='Max Depth for graph generation', dest='max_display_depth')
    group.add_argument('--max-time-cg-gen', help='Max time in seconds to wait for callgraph gen.', default=5)
    group.add_argument('--cg-direction', help='Direction for callgraph.',
                       choices=['calling', 'called', 'both'], default='calling')



class CallGraph:

    def __init__(self, root=None):
        self.graph = {}
        self.title = None
        self.count = 0
        self.max_depth = 0
        self.root = root

    def set_root(self, root: str):
        self.graph.setdefault(root, [])
        self.root = root

    def add_edge(self, node1, node2, depth):

        assert self.root is not None, 'root node must be set prior to adding an edge'

        self.graph.setdefault(node1, [])
        self.graph.setdefault(node2, [])

        self.graph[node1].append((node2, depth, self.count))
        self.count += 1

        # update max depth
        if depth > self.max_depth:
            self.max_depth = depth

    def print_graph(self):
        for src, dst in self.graph.items():
            print(f"{src}-->{dst}")

    def root_at_end(self) -> bool:
        """
        Determines the direction of the graph
        """
        # if the root has no links, the root is at the end
        return len(self.graph[self.root]) == 0

    def get_direction(self):
        """
        reports calling or called if known
        """
        direction = None

        if len(self.graph) == 1:
            direction = 'unknown'
        elif self.root_at_end():
            direction = 'calling'
        else:
            direction = 'called'

        return direction

    def get_endpoints(self) -> list:

        end_nodes = set()

        if not self.root_at_end():
            for src in list(self.graph):
                dst = self.graph[src]
                # special case of loop
                if len(dst) == 0 or len(dst) == 1 and dst[0] == src:
                    end_nodes.add(src)
        else:
            destinations = []

            for src in list(self.graph):
                dst = self.graph[src]
                # special case of loop
                if len(dst) == 1 and dst[0] == src:
                    # don't append to destinations in this case
                    continue

                for d in dst:
                    destinations.append(d[0])

            end_nodes = set(self.graph.keys()).difference(set(destinations))

        return list(end_nodes)

    def get_count_at_depth(self, depth: int) -> int:
        """
        Returns count for nodes at a specific depth
        """

        count = 0
        for src in list(self.graph):
            dst = self.graph[src]
            for d in dst:
                if d[1] == depth:
                    count += 1

        return count

    def links_count(self) -> int:
        """
        Returns count of edges
        """

        count = 0
        for src in list(self.graph):
            dst = self.graph[src]

            for d in dst:
                count += 1

        return count
    
    @staticmethod
    def remove_bad_mermaid_chars(text: str):
        return re.sub(r'`','',text)

    def gen_mermaid_flow_graph(self, direction=None, shaded_nodes: list = None, shade_color='#339933', max_display_depth=None, endpoint_only=False, wrap_mermaid=False) -> str:
        """
        Generate MermaidJS flowchart from self.graph
        See https://mermaid.js.org/syntax/flowchart.html
        """

        # TODO mark root node with circle
        # TODO mark end_points with shape

        # used to create a key index for flowchart ids
        # once defineed, a func name can be represented by a symbol, saving space
        node_keys = {}
        node_count = 0
        existing_base_links = set()

        shade_key = 'sh'

        # use dict to preserve order of links
        links = {}

        # guess best orientation
        if not direction:
            if len(self.graph) < 350:
                direction = 'TD'
            else:
                direction = 'LR'

        mermaid_flow = '''flowchart {direction}\n{style}\n{links}\n'''

        if shaded_nodes:
            style = f'''classDef {shade_key} fill:{shade_color}'''
        else:
            style = ''

        if len(self.graph) == 1:
            links[self.root] = 1
        else:

            if endpoint_only:

                endpoints = self.get_endpoints()

                for i, end in enumerate(endpoints):

                    if shaded_nodes and end in shaded_nodes:
                        end_style_class = f':::{shade_key}'
                    else:
                        end_style_class = ''

                    if shaded_nodes and self.root in shaded_nodes:
                        root_style_class = f':::{shade_key}'
                    else:
                        root_style_class = ''

                    if self.root_at_end():
                        link = f'{i}["{end}"]{end_style_class} --> root["{self.root}"]{root_style_class}'
                    else:
                        link = f'root["{self.root}"]{root_style_class} --> {i}["{end}"]{end_style_class}'

                    links[link] = 1

            else:

                for src in list(self.graph):

                    if shaded_nodes and src in shaded_nodes:
                        src_style_class = f':::{shade_key}'
                    else:
                        src_style_class = ''

                    for node in list(self.graph[src]):

                        depth = node[1]
                        fname = node[0]

                        if max_display_depth is not None and depth > max_display_depth:
                            continue

                        if shaded_nodes and fname in shaded_nodes:
                            dst_style_class = f':::{shade_key}'
                        else:
                            dst_style_class = ''

                        # Build src --> dst link
                        # Don't add duplicate links
                        # Use short ids for func name to save space with node_keys
                        if node_keys.get(src) is None:
                            node_keys[src] = node_count
                            node_count += 1
                            src_node = f'{node_keys[src]}["{src}"]{src_style_class}'
                        else:
                            src_node = f'{node_keys[src]}{src_style_class}'

                        if node_keys.get(fname) is None:
                            node_keys[fname] = node_count
                            node_count += 1
                            dst_node = f'{node_keys[fname]}["{fname}"]{dst_style_class}'
                        else:
                            dst_node = f'{node_keys[fname]}{dst_style_class}'

                        # record base link
                        current_base_link = f'{src} --> {node[0]}'

                        # don't add link if another already exists
                        if not current_base_link in existing_base_links:
                            link = f'{src_node} --> {dst_node}'
                            links[link] = 1
                            existing_base_links.add(current_base_link)
                        # else:
                        #     print('Duplicate base link found!')

        mermaid_chart = mermaid_flow.format(links='\n'.join(links.keys()), direction=direction, style=style)

        mermaid_chart = self.remove_bad_mermaid_chars(mermaid_chart)

        if wrap_mermaid:
            mermaid_chart = _wrap_mermaid(mermaid_chart)

        return mermaid_chart

    def gen_mermaid_mind_map(self, max_display_depth=None, wrap_mermaid=False) -> str:
        """
        Generate MermaidJS mindmap from self.graph
        See https://mermaid.js.org/syntax/mindmap.html
        """

        rows = []

        mermaid_mind = '''mindmap\nroot(({root}))\n{rows}\n'''

        destinations = []

        for src in list(self.graph):
            dst = self.graph[src]
            for d in dst:
                destinations.append(d)

        last_depth = 0
        current_level_names = []
        for i, row in enumerate(sorted(destinations, key=lambda x: x[2])):
            depth = row[1]

            # skip root row
            if depth < 2 or max_display_depth is not None and depth > max_display_depth:
                continue

            if depth < last_depth:
                # reset level names
                current_level_names = []

            if not row[0] in current_level_names:
                spaces = (depth+1)*'  '
                rows.append(f"{spaces}{row[0]}")
                last_depth = depth
                current_level_names.append(row[0])

        mermaid_chart = mermaid_mind.format(rows='\n'.join(rows), root=self.root)

        mermaid_chart = self.remove_bad_mermaid_chars(mermaid_chart)

        if wrap_mermaid:
            mermaid_chart = _wrap_mermaid(mermaid_chart)

        return mermaid_chart


@lru_cache(None)
def get_calling_funcs_memo(f: "ghidra.program.model.listing.Function"):
    return list(f.getCallingFunctions(None))


@lru_cache(None)
def get_called_funcs_memo(f: "ghidra.program.model.listing.Function"):
    return list(f.getCalledFunctions(None))


# Recursively calling to build calling graph
def get_calling(f: "ghidra.program.model.listing.Function", cgraph: CallGraph = None, depth: int = 0, visited: tuple = None, verbose=False, include_ns=True, start_time=None, max_run_time=None, max_depth=MAX_DEPTH):
    """
    Build a call graph of all calling functions
    Traverses depth first
    """

    if f == None:
        return None
    
    if cgraph is None:
        cgraph = CallGraph()

    if depth == 0:
        if verbose:
            print(f"root({f.getName(include_ns)})")
        cgraph.set_root(f.getName(include_ns))
        visited = tuple()
        start_time = time.time()

    if depth == MAX_DEPTH:
        cgraph.add_edge(f.getName(include_ns), f'MAX_DEPTH_HIT - {depth}', depth)
        return cgraph

    if (time.time() - start_time) > float(max_run_time):
        #raise TimeoutError(f'time expired for {clean_func(f,include_ns)}')
        cgraph.add_edge(f.getName(include_ns), f'MAX_TIME_HIT - time: {max_run_time} depth: {depth}', depth)
        print(f'\nWarn: cg : {cgraph.root} edges: {cgraph.links_count()} depth: {depth} name: {f.name} did not complete. max_run_time: {max_run_time} Increase timeout with --max-time-cg-gen MAX_TIME_CG_GEN')
        return cgraph

    space = (depth+2)*'  '

    # loop check
    if f.getName(True) in visited:

        # calling loop
        if verbose:
            print(f"{space} - LOOOOP {f.getName(include_ns)}")

        return cgraph

    calling = get_calling_funcs_memo(f)

    visited = visited + tuple([f.getName(True)])

    if len(calling) > 0:

        depth = depth+1

        for c in calling:

            if verbose:
                print(f"{space} - {c.getName(include_ns)}")

            # Add calling edge
            cgraph.add_edge(c.getName(include_ns), f.getName(include_ns), depth)

            # Parse further functions
            cgraph = get_calling(c, cgraph, depth, visited=visited, start_time=start_time, max_run_time=max_run_time, max_depth=max_depth)
    else:
        if verbose:
            print(f'{space} - END for {f.name}')

    return cgraph


def func_is_external(f: "ghidra.program.model.listing.Function"):
    # sometimwa f.exExternal() failes (like with ls binary)
    return (f.isExternal() or "<EXTERNAL>" in f.getName(True))

# Recursively calling to build called graph


def get_called(f: "ghidra.program.model.listing.Function", cgraph: CallGraph = None, depth: int = 0, visited: list = [], verbose=False, include_ns=True, start_time=None, max_run_time=None, max_depth=MAX_DEPTH):
    """
    Build a call graph of all called functions
    Traverses depth first
    """

    if f == None:
        return None
    
    if cgraph is None:
        cgraph = CallGraph()

    if depth == 0:
        if verbose:
            print(f"root({f.getName(include_ns)})")
        cgraph.set_root(f.getName(include_ns))
        visited = tuple()
        start_time = time.time()

    if depth == max_depth:
        cgraph.add_edge(f.getName(include_ns), f'MAX_DEPTH_HIT - {depth}', depth)
        return cgraph

    if (time.time() - start_time) > float(max_run_time):
        cgraph.add_edge(f.getName(include_ns), f'MAX_TIME_HIT - time: {max_run_time} depth: {depth}', depth)
        print(f'\nWarn: cg : {cgraph.root} edges: {cgraph.links_count()} depth: {depth} name: {f.name} did not complete. max_run_time: {max_run_time} Increase timeout with --max-time-cg-gen MAX_TIME_CG_GEN')
        return cgraph
        
    space = (depth+2)*'  '

    # loop check
    if f.getName(True) in visited:

        # calling loop
        if verbose:
            print(f"{space} - LOOOOP {f.getName(include_ns)}")

        return cgraph

    visited = visited + tuple([f.getName(True)])

    called = get_called_funcs_memo(f)

    if len(called) > 0:

        # this check handles special case when get_called(f) is external but returns called func of itself
        # in that case ignore it
        if not (func_is_external(f) and len(called) == 1):

            depth = depth+1

            for c in called:
                c: "ghidra.program.model.listing.Function" = c

                if verbose:
                    print(f"{space} - {c.getName(include_ns)}")

                # Add called edge
                if func_is_external(c):

                    # force external to show namespace lib with sendind param True
                    cgraph.add_edge(f.getName(include_ns), f"{c.getName(True)}", depth)

                else:
                    cgraph.add_edge(f.getName(include_ns), c.getName(include_ns), depth)

                    # Parse further functions
                    cgraph = get_called(c, cgraph, depth, visited=visited,
                                        start_time=start_time, max_run_time=max_run_time, max_depth=max_depth)

    else:
        if verbose:
            print(f'{space} - END for {f.name}')

    return cgraph


def _wrap_mermaid(text: str) -> str:
    return f'''```mermaid\n{text}\n```'''


def gen_mermaid_url(graph: str, edit=False) -> str:
    """
    Generate valid mermaid live edit and image links
    # based on serialize func  https://github.com/mermaid-js/mermaid-live-editor/blob/b5978e6faf7635e39452855fb4d062d1452ab71b/src/lib/util/serde.ts#L19-L24
    """

    mm_json = {'code': graph, 'mermaid': {'theme': 'dark'}, 'updateEditor': True,
               'autoSync': True, 'updateDiagram': True, "editorMode": "code", "panZoom": True}
    base64_string = base64.urlsafe_b64encode(zlib.compress(json.dumps(mm_json).encode('utf-8'), 9)).decode('ascii')

    if edit:
        url = f'https://mermaid.live/edit#pako:{base64_string}'
    else:
        url = f'https://mermaid.ink/img/svg/pako:{base64_string}'

    return url


def gen_callgraph_md(f: "ghidra.program.model.listing.Function", called: str, calling: str, calling_entrypoints: str, called_endpoints: str, called_mind: str, calling_mind: str):

    fname = f.getName(True)

    calling_mind_url = f'[Edit calling Mindmap]({gen_mermaid_url(calling_mind,edit=True)})'
    called_mind_url = f'![Edit called Mindmap]({gen_mermaid_url(called_mind,edit=True)})'

    md_template = f'''
# {fname}

## Calling

Functions that call `{fname}`.

### Flowchart

[Edit on mermaid live]({gen_mermaid_url(calling,edit=True)})

{_wrap_mermaid(calling)}

### Entrypoints

A condensed view, showing only entrypoints to the callgraph.

{_wrap_mermaid(calling_entrypoints)}

### Mindmap

{calling_mind_url}

## Called

Functions that `{fname}` calls

### Flowchart

[Edit on mermaid live]({gen_mermaid_url(called,edit=True)})

{_wrap_mermaid(called)}

### Endpoints

A condensed view, showing only endpoints of the callgraph.

{_wrap_mermaid(called_endpoints)}

### Mindmap

{called_mind_url}

'''

    return md_template


def gen_callgraph(func: 'ghidra.program.model.listing.Function', max_display_depth=None, direction='calling', max_run_time=None, name=None):

    if name is None:
        name = f'{func.getName()[:MAX_PATH_LEN]}-{func.entryPoint}'    
    
    # print(f'Generating {direction} callgraph for : {name}')
    flow = ''
    callgraph = None

    if direction == 'calling':
        callgraph = get_calling(func, max_run_time=max_run_time)
    elif direction == 'called':
        callgraph = get_called(func, max_run_time=max_run_time)
    else:
        raise Exception(f'Unsupported callgraph direction {direction}')

    if callgraph is not None:
        flow = callgraph.gen_mermaid_flow_graph(
            shaded_nodes=callgraph.get_endpoints(),
            max_display_depth=max_display_depth,
            wrap_mermaid=True)
        flow_ends = callgraph.gen_mermaid_flow_graph(
            shaded_nodes=callgraph.get_endpoints(), endpoint_only=True, wrap_mermaid=True)
        mind = callgraph.gen_mermaid_mind_map(max_display_depth=3, wrap_mermaid=True)

    return [name, direction, callgraph, [['flow', flow], ['flow_ends', flow_ends], ['mind', mind]]]


import argparse
import re
from pathlib import Path

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import ghidra
    from ghidra_builtins import *

def has_bsim() -> bool:

    from ghidra.feature.vt import api as vtapi
    has_bsim = False

    has_bsim = hasattr(vtapi,'BSimProgramCorrelatorFactory')
    
    return has_bsim

def add_bsim_args_to_parser(parser: argparse.ArgumentParser):
    
    group = parser.add_argument_group('BSim Options')
    group.add_argument('--bsim',  help='Generate BSim function feature vector signatures', action='store_true')
    group.add_argument('--bsim-sig-path',  help='Path to store BSim xml sigs', default='bsim-xmls')
    group.add_argument('--bsim-template',  help='BSim database template', default='medium_nosize')
    # group.add_argument('--bsim-db',  help='Path to local BSim database', default=None)
    group.add_argument('--bsim-cat',  help='BSim category. (type:value) --bsim-cat color:red', action='append')
    #group.add_argument('--bsim-ftag',  help='BSim function', action='append')
    
# TODO
# def create_local_h2_db(db_path: Path):
#     pass

# TODO
# def add_program_to_bsim_db(db_path: Path, func_filter: str):
#     pass
    

def add_categories_to_prog(prog,categories) -> list:
    """
    Need to set program option first
    See: Features/BSim/src/main/java/ghidra/features/bsim/query/GenSignatures.java#L231
    """

    from ghidra.program.model.listing import Program

    prog_opts = prog.getOptions(Program.PROGRAM_INFO)
    cat_types = []

    # #debug
    # for name in prog_opts.optionNames:
    #     print(f'{name} : {prog_opts.getDefaultValueAsString(name)}')

    for cat in categories:
        if cat.find(':') == -1:
            # no cat option value defined (assume default prog option)
            cat_type = cat
            val = prog_opts.getDefaultValueAsString(cat_type)
            if not prog_opts.contains(cat_type):
                raise ValueError(f'--bsim-cat arg "{cat_type}" provided is not valid. Custom types need value suppled type:value.')
        else:
            cat_type,val = cat.split(':')
            prog_opts.setString(cat_type, val)

        cat_types.append(cat_type)

        print(f'Adding category type:{cat_type} val:{val}')

    return cat_types


def remove_temp_prog_options(prog,categories):
    """
    remove non permanent options (cats already added)
    this is to avoid cluttering up program options
    """

    from ghidra.program.model.listing import Program

    prog_opts = prog.getOptions(Program.PROGRAM_INFO)

    for cat in categories:
        if cat.find(':') != -1:
            cat_type,val = cat.split(':')
            prog_opts.removeOption(cat_type)
            assert not prog_opts.contains(cat_type)
        else:
            assert prog_opts.contains(cat)

def gen_bsim_sigs_for_program(prog: "ghidra.program.model.listing.Program", sigs_path: Path, bsim_template,categories=None,all_funcs=None):
    """
    Generates signatures for all the functions in a program
    Optionally: geneatures sigs for funcs passed into all_funcs
    """

    if not has_bsim():
        raise Exception('Calling BSim but BSim is not found')

    # see Ghidra/Features/BSim/ghidra_scripts/GenerateSignatures.py

    import java.lang.System as System
    import java.io.File as File
    from java.util import ArrayList
    import ghidra.features.bsim.query.FunctionDatabase as FunctionDatabase
    import ghidra.features.bsim.query.GenSignatures as GenSignatures
    from ghidra.features.bsim.query.description import DescriptionManager

    import java.io.FileWriter as FileWriter
    from ghidra.util.task import ConsoleTaskMonitor
    from ghidra.program.model.address import AddressSet
    from ghidra.framework.protocol.ghidra import GhidraURL

    md5String = prog.getExecutableMD5()
    monitor = ConsoleTaskMonitor().DUMMY_MONITOR
    sigs_path = Path(sigs_path)
    sigs_path.mkdir(exist_ok=True,parents=True)
    cat_add_count = 0
    func_count = 0


    if (md5String is None) or (len(md5String) < 10):
        raise Exception("Could not get MD5 on file: " + prog.getName())
    basename = "sigs_" + md5String + "_" + prog.getName()
    System.setProperty("ghidra.output",basename)
    if not sigs_path.exists():
        raise FileNotFoundError(f'Missing dir {sigs_path}')
    # TODO pull this into a init func for gensig
    vectorFactory = FunctionDatabase.generateLSHVectorFactory()
    gensig : GenSignatures = GenSignatures(True)
    templateName = bsim_template
    config = FunctionDatabase.loadConfigurationTemplate(templateName)
    vectorFactory.set(config.weightfactory, config.idflookup, config.info.settings)
    gensig.setVectorFactory(vectorFactory)
    # this adds from db config
    gensig.addExecutableCategories(config.info.execats)
    gensig.addFunctionTags(config.info.functionTags)
    gensig.addDateColumnName(config.info.dateColumnName)
    
    df = prog.getDomainFile().getParent()
    folderURL = df.getLocalProjectURL()
    normalizedProjectURL = GhidraURL.getProjectURL(folderURL);
    repo = normalizedProjectURL.toExternalForm();
    path = GenSignatures.getPathFromDomainFile(prog)

    # add custom exe categories
    if categories is not None:
        cat_types = add_categories_to_prog(prog, categories)
        cats = ArrayList()
        
        # add cat to list
        for cat_type in cat_types:
            cats.add(cat_type)

        # add arg categories to gensig
        gensig.addExecutableCategories(cats)

    gensig.openProgram(prog,None,None,None,repo,path)

    if categories is not None:
        remove_temp_prog_options(prog,categories)
    
    # Only generate sigs for filter
    if all_funcs is not None:
        # convert functions to Function Iterator
        # Ghidra/Features/Base/ghidra_scripts/SelectFunctionsScript.java#L30
        all_func_addr_set = AddressSet()
        fman = prog.getFunctionManager()

        for func in all_funcs:
            all_func_addr_set.add(func.getBody())
        iter = fman.getFunctions(all_func_addr_set,True)
        gensig.scanFunctions(iter, len(all_funcs), monitor)
    else:
        # generate for all functions
        iter = fman.getFunctions(True)
        gensig.scanFunctions(iter, fman.getFunctionCount(), monitor)
    
    

    # write out sig
    outfile = File(str(sigs_path.absolute()),basename)
    fwrite = FileWriter(outfile)
    manager : DescriptionManager  = gensig.getDescriptionManager()
    func_count = manager.numFunctions()
    # if er.nameExec == prog.getName()
    if categories is not None:
        manager_cats = [(cat.type,cat.category) for er in manager.executableRecordSet if er.allCategories is not None for cat in er.allCategories  ]
        print(f'Categories added: {manager_cats}')
        cat_add_count = len(manager_cats)
        assert cat_add_count == len(categories)
    manager.saveXml(fwrite)
    fwrite.close()

    # close
    gensig.dispose();

    return (basename,func_count, cat_add_count)



import re
from pathlib import Path
from typing import TYPE_CHECKING
from argparse import Namespace
import concurrent.futures
from time import time
import json
import hashlib
from pyghidra import HeadlessPyGhidraLauncher, open_program

from .utility import set_pdb, setup_symbol_server, set_remote_pdbs, analyze_program, get_pdb, apply_gdt, save_program_as_gzf
from .callgraph import get_called, get_calling, CallGraph, gen_callgraph
from .bsim import gen_bsim_sigs_for_program, has_bsim

# needed for ghidra python vscode autocomplete
if TYPE_CHECKING:
    import ghidra
    from ghidra_builtins import *

MAX_PATH_LEN = 50


def get_filename(func: 'ghidra.program.model.listing.Function'):
    return f'{func.getName()[:MAX_PATH_LEN]}-{func.entryPoint}'


def get_md5_file_digest(path: str) -> str:
    # https://stackoverflow.com/questions/22058048/hashing-a-file-in-python
    # BUF_SIZE is totally arbitrary, change for your app!
    BUF_SIZE = 65536  # lets read stuff in 64kb chunks!

    path = Path(path)

    md5 = hashlib.md5()

    with path.open('rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            md5.update(data)

    return f'{md5.hexdigest()}'


def gen_proj_bin_name_from_path(path: Path):
    """
    Generate unique project name from binary for Ghidra Project
    """

    return '-'.join((path.name, get_md5_file_digest(path.absolute())))


def get_bin_output_path(output_path: Path, bin_name: str):

    return Path(output_path) / 'bins' / bin_name


def setup_decompliers(program: "ghidra.program.model.listing.Program", thread_count: int = 2) -> dict:
    """
    Setup decompliers to use during diff bins. Each one must be initialized with a program.
    """

    from ghidra.app.decompiler import DecompInterface
    from ghidra.app.decompiler import DecompileOptions

    decompilers = {}
    decomp_options = DecompileOptions()
    decomp_options.grabFromProgram(program)

    for i in range(thread_count):
        decompilers.setdefault(i, DecompInterface())
        decompilers[i].setOptions(decomp_options)
        decompilers[i].openProgram(program)

    print(f'Setup {thread_count} decompliers')

    return decompilers


def decompile_func(func: 'ghidra.program.model.listing.Function',
                   decompilers: dict,
                   thread_id: int = 0,
                   timeout: int = 0,
                   monitor=None) -> list:
    """
    Decompile function and return [funcname, decompilation]
    Ghidra/Features/Decompiler/src/main/java/ghidra/app/util/exporter/CppExporter.java#L514
    """
    from ghidra.util.task import ConsoleTaskMonitor
    from ghidra.app.decompiler import DecompiledFunction, DecompileResults

    if monitor is None:
        monitor = ConsoleTaskMonitor()

    result: "DecompileResults" = decompilers[thread_id].decompileFunction(func, timeout, monitor)

    if '' == result.getErrorMessage():
        code = result.decompiledFunction.getC()
        sig = result.decompiledFunction.getSignature()
    else:
        code = result.getErrorMessage()
        sig = None

    return [get_filename(func), code, sig]


def decompile_to_single_file(path: Path,
                             prog: "ghidra.program.model.listing.Program",
                             create_header: bool = True,
                             create_file: bool = True,
                             emit_types: bool = True,
                             exclude_tags: bool = False,
                             tags: str = None,
                             verbose: bool = True):
    """
    Use Ghidra's CppExporter to decompile all functions to a single file
    """
    from ghidra.app.util.exporter import CppExporter
    from ghidra.util.task import ConsoleTaskMonitor
    from java.io import File

    c_file = File(path.absolute())

    if verbose:
        monitor = ConsoleTaskMonitor()
    else:
        monitor = ConsoleTaskMonitor().DUMMY

    try:
        # Ghidra CppExporter before 10.3.3 and later
        decompiler = CppExporter(None, create_header, create_file, emit_types, exclude_tags, tags)
    except TypeError:
        # Ghidra CppExporter before 10.3.3
        decompiler = CppExporter(create_header, create_file, emit_types, exclude_tags, tags)

    decompiler.export(c_file, prog, prog.getMemory(), monitor)


def decompile(args: Namespace):

    print(f'Starting decompliations: {args}')

    bin_path = Path(args.bin)
    bin_proj_name = gen_proj_bin_name_from_path(bin_path)
    thread_count = args.thread_count

    output_path = Path(args.output_path)
    bin_output_path = get_bin_output_path(output_path, bin_proj_name)
    decomp_path = bin_output_path / 'decomps'
    output_path.mkdir(exist_ok=True, parents=True)
    bin_output_path.mkdir(exist_ok=True, parents=True)
    decomp_path.mkdir(exist_ok=True, parents=True)

    if args.project_path == 'ghidra_projects':
        project_location = output_path / args.project_path
    else:
        project_location = Path(args.project_path)

    gzf_path = None
    if args.gzf:
        if args.gzf_path == 'gzfs':
            gzf_path = output_path / args.gzf_path
        else:
            gzf_path = Path(args.gzf_path)

        gzf_path.mkdir(exist_ok=True, parents=True)

    if args.symbols_path == 'symbols':
        symbols_path = output_path / args.symbols_path
    else:
        symbols_path = Path(args.symbols_path)

    if args.bsim_sig_path == 'bsim_xmls':
        bsim_sig_path = output_path / args.bsim_sig_path
    else:
        bsim_sig_path = output_path / Path(args.bsim_sig_path)

    # turn on verbose
    launcher = HeadlessPyGhidraLauncher(True)

    # set max % of host RAM
    launcher.add_vmargs(f'-XX:MaxRAMPercentage={args.max_ram_percent}')
    if args.print_flags:
        launcher.add_vmargs('-XX:+PrintFlagsFinal')

    launcher.start()

    from ghidra.util.task import ConsoleTaskMonitor
    monitor = ConsoleTaskMonitor()
    from ghidra.program.model.listing import Program

    # Setup and analyze project
    with open_program(bin_path, project_location=project_location, project_name=bin_proj_name, analyze=False) as flat_api:

        program: "Program" = flat_api.getCurrentProgram()

        if not args.skip_symbols:
            if args.sym_file_path:
                set_pdb(program, args.sym_file_path)
            else:
                setup_symbol_server(symbols_path)

                set_remote_pdbs(program, True)

            pdb = get_pdb(program)
            if pdb is None:
                print(f"Failed to find pdb for {program}")

        # apply GDT
        if args.gdt:
            for gdt_path in args.gdt:
                print(f'Applying gdt {gdt_path}...')
                apply_gdt(program, gdt_path, verbose=args.va)

        gdt_names = [name for name in program.getDataTypeManager().getSourceArchives()]
        if len(gdt_names) > 0:
            print(f'Using file gdts: {gdt_names}')

        # analyze program if we haven't yet
        analyze_program(program, verbose=args.va, force_analysis=args.fa, gzf_path=gzf_path)

    # Save copy of program in gzf after analysis
    if args.gzf:
        from ghidra.base.project import GhidraProject
        try:
            project = GhidraProject.openProject(Path(project_location / bin_proj_name), bin_proj_name, True)
            program = project.openProgram("/", bin_path.name, True)
            save_program_as_gzf(program, gzf_path / bin_proj_name, project)
        finally:
            project.close(program)
            project.close()

    # decompile and callgraph all the things
    with open_program(bin_path, project_location=project_location, project_name=bin_proj_name, analyze=False) as flat_api:

        all_funcs = []
        skip_count = 0

        program: "Program" = flat_api.getCurrentProgram()

        for f in program.functionManager.getFunctions(True):

            if args.filters:
                if any([re.search(fil, f.name, re.IGNORECASE) for fil in args.filters]):
                    all_funcs.append(f)
                else:
                    skip_count += 1
            else:
                all_funcs.append(f)

        if skip_count > 0:
            print(f'Skipped {skip_count} functions that failed to match any of {args.filters}')

        decompilations = []
        callgraphs = []

        if args.cppexport:
            print(f"Decompiling {len(all_funcs)} functions using Ghidra's CppExporter")
            c_file = decomp_path / Path(bin_path.name + '.c')
            start = time()
            decompile_to_single_file(c_file, program)
            print(f'Decompiled {len(all_funcs)} functions for {program.name} in {time() - start}')
            print(f"Wrote results to {c_file} and {c_file.stem + '.h'}")
        else:
            print(f'Decompiling {len(all_funcs)} functions using {thread_count} threads')

            decompilers = setup_decompliers(program, thread_count)
            completed = 0

            # Decompile all files
            start = time()
            with concurrent.futures.ThreadPoolExecutor(max_workers=thread_count) as executor:
                futures = (executor.submit(decompile_func, func, decompilers, thread_id % thread_count, monitor=monitor)
                           for thread_id, func in enumerate(all_funcs) if args.skip_cache or not (decomp_path / (get_filename(func) + '.c')).exists())

                for future in concurrent.futures.as_completed(futures):
                    decompilations.append(future.result())
                    completed += 1
                    if (completed % 100) == 0:
                        print(f'Decompiled {completed} and {int(completed/len(all_funcs)*100)}%')

            print(f'Decompiled {completed} functions for {program.name} in {time() - start}')
            print(f'{len(all_funcs) - completed} decompilations already existed.')

            # Save all decomps
            start = time()
            with concurrent.futures.ThreadPoolExecutor(max_workers=thread_count) as executor:
                futures = (executor.submit((decomp_path / (name + '.c')).write_text, decomp)
                           for name, decomp, sig in decompilations)

                for future in concurrent.futures.as_completed(futures):
                    pass

            print(f'Wrote {completed} decompilations for {program.name} to {decomp_path} in {time() - start}')

            # Generate callgrpahs for functions
            if args.callgraphs:

                start = time()
                completed = 0
                callgraph_path = bin_output_path / 'callgraphs'
                callgraphs_completed_path = callgraph_path / 'completed_callgraphs.json'
                if callgraphs_completed_path.exists():
                    callgraphs_completed = json.loads(callgraphs_completed_path.read_text())
                else:
                    callgraphs_completed = []

                callgraph_path.mkdir(exist_ok=True)

                if args.cg_direction == 'both':
                    directions = ['called', 'calling']
                else:
                    directions = [args.cg_direction]

                max_display_depth = None
                if args.max_display_depth is not None:
                    max_display_depth = int(args.max_display_depth)

                with concurrent.futures.ThreadPoolExecutor(max_workers=thread_count) as executor:
                    futures = (executor.submit(gen_callgraph, func, max_display_depth, direction, args.max_time_cg_gen, get_filename(func))
                               for direction in directions for func in all_funcs if args.skip_cache or get_filename(func) not in callgraphs_completed and re.search(args.callgraph_filter, func.name) is not None)

                    for future in concurrent.futures.as_completed(futures):

                        callgraphs.append(future.result())
                        name, direction, callgraph, graphs = callgraphs[-1]

                        for ctype, chart in graphs:
                            (callgraph_path / (name + f'.{ctype}.{direction}.md')).write_text(chart)
                        callgraphs_completed.append(name)

                        completed += 1
                        if (completed % 100) == 0:
                            per_complete = int(completed/len(all_funcs)*100*len(directions))
                            print(f'\nGenerated callgraph {completed} and {per_complete}%\n')

                callgraphs_completed_path.write_text(json.dumps(callgraphs_completed))
                print(f'Callgraphed {completed} functions for {program.name} in {time() - start}')
                print(f'Wrote {completed} callgraphs for {program.name} to {callgraph_path} in {time() - start}')
                print(f'{len(all_funcs) - completed} callgraphs already existed.')

        # BSim
        gensig = None
        manager = None
        if args.bsim:

            if has_bsim():
                start = time()
                print(f'Generating BSim sigs for {len(all_funcs)} functions for {program.name}')
                sig_name, func_count, cat_count = gen_bsim_sigs_for_program(
                    program, bsim_sig_path, args.bsim_template, args.bsim_cat, all_funcs)
                print(f'Generated BSim sigs for {func_count} functions in {time() - start}')
                print(f'Sigs are in {bsim_sig_path / sig_name}')
            else:
                print('WARN: Skipping BSim. BSim not present')

        return (all_funcs, decompilations, bin_output_path, str(program.compiler), str(program.languageID), callgraphs)




from pathlib import Path
from typing import Union, TYPE_CHECKING

# needed for ghidra python vscode autocomplete
if TYPE_CHECKING:
    import ghidra
    from ghidra_builtins import *  # noqa: F403


def analyze_program(program: "ghidra.program.model.listing.Program", verbose: bool = False, force_analysis: bool = False, save: bool = False, gzf_path: Path = None):
    """
    Modified pyghidra.core._analyze_program    
    """

    from ghidra.program.flatapi import FlatProgramAPI
    from ghidra.util.task import ConsoleTaskMonitor
    from ghidra.program.util import GhidraProgramUtilities
    from ghidra.app.script import GhidraScriptUtil

    if verbose:
        print('Enabling verbose analysis...')
        monitor = ConsoleTaskMonitor()
        flat_api = FlatProgramAPI(program, monitor)
    else:
        flat_api = FlatProgramAPI(program)

    if GhidraProgramUtilities.shouldAskToAnalyze(program) or force_analysis:
        print(f"Analyzing program {program.name}...")

        GhidraScriptUtil.acquireBundleHostReference()
        try:
            print(f"Running analyzers...")
            flat_api.analyzeAll(program)
            if hasattr(GhidraProgramUtilities, 'setAnalyzedFlag'):
                GhidraProgramUtilities.setAnalyzedFlag(program, True)
            elif hasattr(GhidraProgramUtilities, 'markProgramAnalyzed'):
                GhidraProgramUtilities.markProgramAnalyzed(program)
            else:
                raise Exception('Missing set analyzed flag method!')

            if save:
                flat_api.saveProgram(program)
        finally:
            GhidraScriptUtil.releaseBundleHostReference()
    else:
        print(f'{program} already analyzed... skipping')


def save_program_as_gzf(program: "ghidra.program.model.listing.Program", gzf_path: Path, project):
    from java.io import File
    from ghidra.base.project import GhidraProject
    # from java.io import IOException
    print(f'Saving gzf archive to {gzf_path}.gzf')

    # GhidraProject.saveAsPackedFile(program, File(f'{gzf_path.absolute()},{program.name}.gzf'), True)
    from java.io import File
    # project.close()
    project.saveAsPackedFile(program, File(f'{gzf_path}.gzf'), True)


def setup_symbol_server(symbols_path: Union[str, Path], level=1, server_urls=None) -> None:
    """
    setup symbols to allow Ghidra to download as needed
    1. Configures symbol_path as local symbol store path
    2. Sets Index level for local symbol path
    - Level 0 indexLevel is a special Ghidra construct - plain directory with a collection of Pdb files
    - Level 1, with pdb files stored directly underthe root directory
    - Level 2, using the first 2 characters of the pdb filename as a bucket to place each pdb file-directory in
    [symbol-store-folder-tree](https://learn.microsoft.com/en-us/windows-hardware/drivers/debugger/symbol-store-folder-tree)
    """

    from pdb_ import PdbPlugin
    from pdb_.symbolserver import LocalSymbolStore
    from pdb_.symbolserver import HttpSymbolServer
    from pdb_.symbolserver import SymbolServerService
    from ghidra.framework import Application

    from java.io import File
    from java.net import URI
    from java.util import ArrayList

    print("Setting up Symbol Server for symbols...")
    print(f"path: {symbols_path} level: {level}")

    symbols_path = Path(symbols_path).absolute()

    # Configure local symbols directory
    symbolsDir = File(symbols_path)
    localSymbolStore = LocalSymbolStore(symbols_path)

    # Create local symbol server
    # pdb/symbolserver/LocalSymbolStore.java#L67
    localSymbolStore.create(symbolsDir, level)

    # Configure well known symbol urls
    if server_urls is None:
        # load wellknown servers
        # Ghidra/Features/PDB/src/main/java/pdb/symbolserver/ui/WellKnownSymbolServerLocation.java#L89
        known_urls = []
        pdbUrlFiles = Application.findFilesByExtensionInApplication(".pdburl")
        for pdbFile in pdbUrlFiles:
            data = Path(pdbFile.absolutePath).read_text()
            print(f"Loaded well known {pdbFile.absolutePath}' length: {len(data)}'")
            for line in data.splitlines(True):
                cat, location, warning = line.split('|')
                known_urls.append(location)
        server_urls = known_urls
    else:
        if not isinstance(server_urls, list):
            raise TypeError('server_urls must be a list of urls')

    symServers = ArrayList()

    for url in server_urls:
        symServers.add(HttpSymbolServer(URI.create(url)))

    symbolServerService = SymbolServerService(localSymbolStore, symServers)

    PdbPlugin.saveSymbolServerServiceConfig(symbolServerService)

    print(f'Symbol Server Configured path: {symbolServerService.toString().strip()}')


def get_pdb(prog: "ghidra.program.model.listing.Program") -> "java.io.File":
    """
    Searches the currently configured symbol server paths for a Pdb symbol file.
    """

    from pdb_.symbolserver import FindOption
    from ghidra.util.task import ConsoleTaskMonitor
    from pdb_ import PdbPlugin

    if hasattr(FindOption, 'ALLOW_UNTRUSTED'):
        # Ghidra 11.2 +
        find_opts = FindOption.of(FindOption.ALLOW_UNTRUSTED)
    else:
        # Ghidra < 11.2
        find_opts = FindOption.of(FindOption.ALLOW_REMOTE)

    # Ghidra/Features/PDB/src/main/java/pdb/PdbPlugin.java#L191
    pdb = PdbPlugin.findPdb(prog, find_opts, ConsoleTaskMonitor())

    return pdb


def set_pdb(program: "ghidra.program.model.listing.Program", path: Union[str, Path]):

    from ghidra.app.plugin.core.analysis import PdbUniversalAnalyzer
    from java.io import File

    print(f'Setting pdb to {symbol_path}')
    symbol_path = Path(path)
    pdbFile = File(symbol_path)
    PdbUniversalAnalyzer.setPdbFileOption(program, pdbFile)


def set_remote_pdbs(program: "ghidra.program.model.listing.Program", allow: bool):
    """
    Enable or disable remote PDB downloads
    """

    from ghidra.app.plugin.core.analysis import PdbAnalyzer
    from ghidra.app.plugin.core.analysis import PdbUniversalAnalyzer
    # Enable Remote Symbol Servers

    if hasattr(PdbUniversalAnalyzer, 'setAllowUntrustedOption'):
        # Ghidra 11.2 +
        PdbUniversalAnalyzer.setAllowUntrustedOption(program, True)
        PdbAnalyzer.setAllowUntrustedOption(program, True)
    else:
        # Ghidra < 11.2
        PdbUniversalAnalyzer.setAllowRemoteOption(program, True)
        PdbAnalyzer.setAllowRemoteOption(program, True)


def apply_gdt(program: "ghidra.program.model.listing.Program", gdt_path:  Union[str, Path], verbose: bool = False):
    """
    Apply GDT to program
    """

    from ghidra.app.cmd.function import ApplyFunctionDataTypesCmd
    from ghidra.program.model.symbol import SourceType
    from java.io import File
    from java.util import List
    from ghidra.program.model.data import FileDataTypeManager
    from ghidra.util.task import ConsoleTaskMonitor

    gdt_path = Path(gdt_path)

    if verbose:
        print('Enabling verbose gdt..')
        monitor = ConsoleTaskMonitor()
    else:
        monitor = ConsoleTaskMonitor().DUMMY_MONITOR

    archiveGDT = File(gdt_path)
    archiveDTM = FileDataTypeManager.openFileArchive(archiveGDT, False)
    always_replace = True
    createBookmarksEnabled = True
    cmd = ApplyFunctionDataTypesCmd(List.of(archiveDTM), None, SourceType.USER_DEFINED,
                                    always_replace, createBookmarksEnabled)
    cmd.applyTo(program, monitor)



import pytest
from pathlib import Path

import pyghidra
from ghidrecomp import decompile, get_parser
from ghidrecomp.decompile import get_bin_output_path, gen_proj_bin_name_from_path
from ghidrecomp.bsim import has_bsim, add_bsim_args_to_parser, add_categories_to_prog
from pyghidra import HeadlessPyGhidraLauncher

# check BSIM exists


def test_bsim_should_exist():

    pyghidra.start()

    # check ghidra version and bail if old
    if HeadlessPyGhidraLauncher().app_info.version < '11.0':
        assert not has_bsim()
    else:
        assert has_bsim()


def test_bsim_bad_known_arg(shared_datadir: Path):

    parser = get_parser()

    bin_path = shared_datadir / 'ls_aarch64'

    args = parser.parse_args([f"{bin_path.absolute()}",
                              "--bsim-cat",
                              "CustomArgNeedsValue",
                              "--bsim"
                              ])

    if has_bsim():
        # will only parse args is has_bsim
        with pytest.raises(ValueError):

            all_funcs, decompilations, output_path, compiler, lang_id, callgraphs = decompile(args)

        x = 1 / 1

# check category + func count


def test_bsim_args_with_ls(shared_datadir: Path):

    parser = get_parser()

    bin_path = shared_datadir / 'ls_aarch64'

    args = parser.parse_args([f"{bin_path.absolute()}",
                              "--skip-cache",
                              "--bsim-cat",
                              "Executable Location",
                              "--bsim-cat",
                              "source:ghidrecomp",
                              "--bsim"
                              ])

    bin_proj_name = gen_proj_bin_name_from_path(bin_path)
    expected_output_path = get_bin_output_path(args.output_path, bin_proj_name)

    all_funcs, decompilations, output_path, compiler, lang_id, callgraphs = decompile(args)

    assert len(all_funcs) == 532
    assert len(decompilations) == 532
    assert output_path == expected_output_path
    assert compiler == 'unknown'
    assert lang_id == 'AARCH64:LE:64:v8A'
    assert len(callgraphs) == 0



import pytest
from pathlib import Path

import pyghidra
from ghidrecomp import decompile, get_parser
from ghidrecomp.decompile import get_bin_output_path, gen_proj_bin_name_from_path
from ghidrecomp.bsim import has_bsim, add_bsim_args_to_parser, add_categories_to_prog
from pyghidra import HeadlessPyGhidraLauncher
from pathlib import Path


def test_gzf_created(shared_datadir: Path):

    parser = get_parser()

    bin_path = shared_datadir / 'ls_aarch64'

    args = parser.parse_args([f"{bin_path.absolute()}", "--filter", "ctype", "--skip-cache", "--gzf"])

    bin_proj_name = gen_proj_bin_name_from_path(bin_path)
    expected_output_path = get_bin_output_path(args.output_path, bin_proj_name)

    gzf_output_path = Path(args.output_path) / 'gzfs' / f"{bin_proj_name}.gzf"

    all_funcs, decompilations, output_path, compiler, lang_id, callgraphs = decompile(args)

    assert len(all_funcs) == 8
    assert len(decompilations) == 8
    assert len(callgraphs) == 0
    assert expected_output_path == output_path
    assert gzf_output_path.exists()


def test_gzf_created_with_path(shared_datadir: Path):

    parser = get_parser()

    bin_path = shared_datadir / 'ls_aarch64'
    gzf_custom_path = shared_datadir / "custom_gzf"

    args = parser.parse_args([f"{bin_path.absolute()}", "--filter", "ctype", "--skip-cache",
                              "--gzf", "--gzf-path", str(gzf_custom_path)])

    bin_proj_name = gen_proj_bin_name_from_path(bin_path)
    expected_output_path = get_bin_output_path(args.output_path, bin_proj_name)

    gzf_output_path = gzf_custom_path / f"{bin_proj_name}.gzf"

    all_funcs, decompilations, output_path, compiler, lang_id, callgraphs = decompile(args)

    assert len(all_funcs) == 8
    assert len(decompilations) == 8
    assert len(callgraphs) == 0
    assert expected_output_path == output_path
    assert gzf_custom_path.exists()



import argparse
import multiprocessing

from ghidrecomp import __version__

from .callgraph import add_cg_args_to_parser
from .bsim import add_bsim_args_to_parser

THREAD_COUNT = multiprocessing.cpu_count()


def get_parser() -> argparse.ArgumentParser:

    parser = argparse.ArgumentParser(description='ghidrecomp - A Command Line Ghidra Decompiler',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('bin', help='Path to binary used for analysis')
    parser.add_argument('--cppexport', action='store_true', help='Use Ghidras CppExporter to decompile to single file')
    parser.add_argument('--filter', dest='filters', action='append', help='Regex match for function name')
    parser.add_argument('--project-path', help='Path to base ghidra projects ', default='ghidra_projects')
    parser.add_argument('--gzf', help='Export gzf of analyzed project', action='store_true')
    parser.add_argument('--gzf-path', help='Path to store gzf of analyzed project', default='gzfs')
    parser.add_argument('--gdt', help='Additional GDT to apply', nargs='?', action='append')
    parser.add_argument('-o', '--output-path', help='Location for all decompilations', default='ghidrecomps')
    parser.add_argument("-v", "--version", action="version", version=__version__)
    parser.add_argument("--skip-cache", action='store_true',
                        help='Skip cached and genearate new decomp and callgraphs.')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--sym-file-path', help='Specify single pdb symbol file for bin')
    group.add_argument('-s', '--symbols-path', help='Path for local symbols directory', default='symbols')
    group.add_argument('--skip-symbols', help='Do not apply symbols', action='store_true')

    parser.add_argument('-t', '--thread-count', type=int,
                        help='Threads to use for processing. Defaults to cpu count', default=THREAD_COUNT)
    parser.add_argument('--va', help='Enable verbose analysis', action='store_true')
    parser.add_argument('--fa', help='Force new analysis (even if already analyzed)', action='store_true')

    group = parser.add_argument_group('JVM Options')
    group.add_argument('--max-ram-percent', help='Set JVM Max Ram %% of host RAM', default=50.0)
    group.add_argument('--print-flags', help='Print JVM flags at start', action='store_true')

    add_cg_args_to_parser(parser)

    add_bsim_args_to_parser(parser)

    return parser
